import click

from fhir_aggregator import list_all_servers
from fhir_aggregator.cli.click_util import NaturalOrderGroup


@click.group(name='servers', cls=NaturalOrderGroup)
def cli():
    """Server commands."""
    pass


@cli.command(name='list')
def _list():
    """
    Lists all known servers.
    """
    results = list_all_servers()
    # Process or display the results
    click.echo(f'Servers: {results}')





import json


from fhir_aggregator.client import ResearchStudySummary


def test_dbgap_research_study():
    """Query dbGap server."""
    from pprint import pprint
    with open('tests/fixtures/dbgap_research_study.json') as fp:
        research_study = json.load(fp)

        assert research_study
        assert isinstance(research_study, dict)
        dbgap_research_study_summary = ResearchStudySummary(research_study=research_study)
        pprint(dbgap_research_study_summary.research_study['extension'])
        assert dbgap_research_study_summary
        extensions = dbgap_research_study_summary.extensions
        assert len(extensions) == 17
        pprint(extensions)
        assert set(extensions) == set(
            {'research_study_citers_citer_title': 'Integrative genomic analysis of the '
                                                  'human immune response to influenza '
                                                  'vaccination',
             'research_study_citers_citer_url': 'https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3713456',
             'research_study_computed_ancestry_ancestry_count_ancestry': 'LEN',
             'research_study_computed_ancestry_ancestry_count_count': 7,
             'research_study_content_num_analyses': 0,
             'research_study_content_num_documents': 0,
             'research_study_content_num_molecular_datasets': 1,
             'research_study_content_num_phenotype_datasets': 3,
             'research_study_content_num_samples': 246,
             'research_study_content_num_sub_studies': 0,
             'research_study_content_num_subjects': 246,
             'research_study_content_num_variables': 30,
             'research_study_molecular_data_types_molecular_data_type': 'SNP Genotypes '
                                                                        '(Array)',
             'research_study_release_date': '2017-06-02',
             'research_study_study_consents_study_consent': 'GRU-IRB',
             'research_study_study_markersets_study_markerset': 'HumanOmniExpress-12v1_A',
             'research_study_study_overview_url': 'https://www.ncbi.nlm.nih.gov/projects/gap/cgi-bin/study.cgi?study_id=phs000635.v1.p1'}
        )
        print(dbgap_research_study_summary.simplified)
        assert dbgap_research_study_summary.simplified == {'identifier': 'phs000635.v1.p1',
                                                           'resourceType': 'ResearchStudy', 'id': 'phs000635',
                                                           'title': 'Adult Influenza Vaccine Genetics',
                                                           'status': 'completed',
                                                           'description': '\nThis study includes two cohorts composed of healthy adults immunized with seasonal influenza vaccine. The first cohort was composed exclusively of males and second of females. Enrollment was restricted to individuals of European heritage to simplify the genetic analysis. Serum samples were collected prior to immunization and at 14 days and 28 days post immunization. Anti-hemagglutinin and neutralizing antibody activity was measured at each of the three time points for Influenza H1N1, H3N2, and Influenza B strains corresponding to the content of the vaccine. Peripheral blood RNA (PaxGene) were collected before immunization and on days 1, 3, and 14 post immunization. The RNA samples were analyzed using Illumina Human HT-12 v3 (Cohort 1) and Illumina HT-12 v4 beadarrays (Cohort 2).\n',
                                                           'meta': 'U', 'category': 'Prospective Longitudinal Cohort',
                                                           'focus': 'D007251', 'condition': 'C0021400',
                                                           'keyword': 'C0021400',
                                                           'research_study_study_overview_url': 'https://www.ncbi.nlm.nih.gov/projects/gap/cgi-bin/study.cgi?study_id=phs000635.v1.p1',
                                                           'research_study_release_date': '2017-06-02',
                                                           'research_study_study_consents_study_consent': 'GRU-IRB',
                                                           'research_study_content_num_phenotype_datasets': 3,
                                                           'research_study_content_num_molecular_datasets': 1,
                                                           'research_study_content_num_variables': 30,
                                                           'research_study_content_num_documents': 0,
                                                           'research_study_content_num_analyses': 0,
                                                           'research_study_content_num_subjects': 246,
                                                           'research_study_content_num_samples': 246,
                                                           'research_study_content_num_sub_studies': 0,
                                                           'research_study_computed_ancestry_ancestry_count_ancestry': 'LEN',
                                                           'research_study_computed_ancestry_ancestry_count_count': 7,
                                                           'research_study_study_markersets_study_markerset': 'HumanOmniExpress-12v1_A',
                                                           'research_study_molecular_data_types_molecular_data_type': 'SNP Genotypes (Array)',
                                                           'research_study_citers_citer_title': 'Integrative genomic analysis of the human immune response to influenza vaccination',
                                                           'research_study_citers_citer_url': 'https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3713456'}


def test_kf_research_study():
    """Query kf server."""
    with open('tests/fixtures/kf_research_study.json') as fp:
        research_study = json.load(fp)
        assert research_study
        assert isinstance(research_study, dict)
        kf_research_study_summary = ResearchStudySummary(research_study=research_study)
        assert kf_research_study_summary
        extensions = kf_research_study_summary.extensions
        print(extensions)
        assert len(extensions) == 0
        print(kf_research_study_summary.simplified)
        assert kf_research_study_summary.simplified == {'identifier': 'SD_65064P2Z', 'resourceType': 'ResearchStudy',
                                                        'id': '1883519',
                                                        'title': 'INCLUDE: (Sherman) Genomic Analysis of Congenital Heart Defects and Acute Lymphoblastic Leukemia in Children with Down Syndrome',
                                                        'status': 'completed', 'meta': 'SD_65064P2Z',
                                                        'category': '276720006', 'keyword': 'SD65064P2Z'}


def test_gtex_research_study():
    """Query kf server."""
    with open('tests/fixtures/gtex_research_study.json') as fp:
        research_study = json.load(fp)
        # mocker.retrieve = lambda queries: _
        #
        # research_study = mocker.retrieve(['ResearchStudy'])
        assert research_study
        assert isinstance(research_study, dict)
        print(research_study)
        gtex_research_study_summary = ResearchStudySummary(research_study=research_study)
        assert gtex_research_study_summary
        assert gtex_research_study_summary.research_study == research_study
        print(gtex_research_study_summary.research_study)
        extensions = gtex_research_study_summary.extensions
        print(extensions)
        assert len(extensions) == 0
        print(gtex_research_study_summary.simplified)
        assert gtex_research_study_summary.simplified == {'identifier': 'GTEx', 'id': 'f6c7b779-657e-4c89-9a52-d7a39f3c661d', 'resourceType': 'ResearchStudy', 'status': 'completed', 'keyword': 'GTEx', 'meta': 'GTEx'}



import json

import click

from fhir_aggregator import normalize
from fhir_aggregator.cli.click_util import NaturalOrderGroup


@click.group(name='normalize', cls=NaturalOrderGroup)
def cli():
    """Normalization  commands."""
    pass


@cli.command()
@click.argument('results_file', type=click.Path(exists=True))
def results(results_file):
    """
    Normalizes search results stored in a file.

    :param results_file: Path to the file containing raw search results.
    """
    with open(results_file, 'r') as file:
        search_results = json.load(file)

    # normalize is a function that normalizes the search results
    normalized_results = normalize(search_results)

    # For demonstration, let's just print the normalized results
    # In practice, you might save this back to a file or database
    click.echo(json.dumps(normalized_results, indent=4))



import os
from collections import OrderedDict

import click
import yaml
from pathlib import Path
from dataclasses import dataclass, field


@dataclass
class Credential:
    """
    Represents credentials for accessing a service.

    Attributes:
        endpoint (str): The service endpoint URL. Required.
        token (str): Authentication token. Optional.
        username (str): Username for authentication. Optional.
        password (str): Password for authentication. Optional.
    """
    endpoint: str
    token: str = field(default=None)
    username: str = field(default=None)
    password: str = field(default=None)

    def validate(self):
        """
        Validates the Credential instance to ensure it meets the requirements.

        Raises:
            ValueError: If the endpoint is None or an empty string, or if both token and username/password are not provided.
        """
        if not self.endpoint:
            raise ValueError("Endpoint is required.")
        if not self.token and (not self.username or not self.password):
            raise ValueError("Either token or both username and password must be provided.")

    def __post_init__(self):
        """
        When defined on the class, it will be called by the generated __init__()
        """
        self.validate()


class CredentialManager:
    """
    Manages credentials stored in a YAML file, allowing CRUD operations.

    Attributes:
        credentials_file (Path): Path to the YAML file storing credentials.
        credentials (dict): Loaded credentials from the YAML file.
    """

    def __init__(self):
        """
        Initializes the CredentialManager, loading or creating the credentials file.
        """
        self.credentials_file = Path.home() / '.fhir_aggregator/credentials.yaml'
        self.credentials_file.parent.mkdir(parents=True, exist_ok=True)
        self.credentials = self._load_credentials()

    def _load_credentials(self):
        """
        Loads credentials from the YAML file, or initializes an empty dictionary if the file does not exist.

        Returns:
            dict: The loaded or initialized credentials.
        """
        if self.credentials_file.exists():
            with open(self.credentials_file, 'r') as file:
                return yaml.safe_load(file) or {}
        else:
            return {}

    def save_credentials(self):
        """
        Saves the current state of credentials back to the YAML file.
        """
        with open(self.credentials_file, 'w') as file:
            yaml.safe_dump(self.credentials, file)

    def add_or_update_credential(self, credential):
        """
        Adds a new credential or updates an existing one using the endpoint as the key.

        Args:
            credential (Credential): The credential to add or update.
        """
        self.credentials[credential.endpoint] = {
            'token': credential.token,
            'username': credential.username,
            'password': credential.password
        }
        self.save_credentials()

    def delete_credential(self, endpoint):
        """
        Deletes a credential by its endpoint.

        Args:
            endpoint (str): The endpoint of the credential to delete.
        """
        if endpoint in self.credentials:
            del self.credentials[endpoint]
            self.save_credentials()

    def get_credential(self, endpoint):
        """
        Retrieves a credential by its endpoint.

        Args:
            endpoint (str): The endpoint of the credential to retrieve.

        Returns:
            dict: The credential, if found; otherwise, None.
        """
        return self.credentials.get(endpoint, None)


def metrics(search_results) -> dict:
    """
    Calculate metrics from search results.
    :param search_results:
    :return:
    """
    # Placeholder for metrics calculation logic
    # This should be replaced with actual logic to calculate metrics from search_results
    return {
        'total_results': len(search_results),
        # Add more metrics calculation here
    }


def search_all_servers(query) -> dict:
    """
    Search all servers for a given query.
    :param query:
    :return: Bundle
    """
    # Placeholder for search logic
    # This function should return the search results for a given query
    return f"Search results for {query}"


def normalize(results: dict) -> dict:
    # Placeholder for normalization logic
    # This should be replaced with actual logic to normalize search_results
    return results  # Replace this with actual normalization logic


def list_all_servers() -> list[str]:
    """
    Lists all known servers.
    :return:  A list of known server urls.
    """
    # Placeholder for server list
    return []



import json

import click

from fhir_aggregator import metrics
from fhir_aggregator.cli.click_util import NaturalOrderGroup


@click.group(name='metrics', cls=NaturalOrderGroup)
def cli():
    """Metrics  commands."""
    pass


@cli.command()
@click.argument('results_file', type=click.Path(exists=True))
def create(results_file):
    """
    Creates metrics from search results stored in a file.

    :param results_file: Path to the file containing search results.
    """
    with open(results_file, 'r') as file:
        search_results = json.load(file)

    _ = metrics(search_results)

    # Display or store the metrics
    click.echo(f'Metrics: {_}')



import itertools
from collections import Counter

from functional import seq


def identifier(resource):
    """Return the official identifier, or first of a resource."""
    identifiers = resource.get('identifier', [])
    official_identifiers = [_ for _ in identifiers if _.get('use', '') == 'official']
    if not official_identifiers and identifiers:
        return identifiers[0]
    elif official_identifiers:
        return official_identifiers[0]
    else:
        return {}


def retrieve_and_process(server, queries):
    """Query server with queries, collect responses and run through two pipelines."""
    copy1, copy2 = server.retrieve(queries, chain=lambda responses: itertools.tee(itertools.chain.from_iterable(responses), 2))

    items_with_identifiers = seq(copy1).map(lambda _: (_['resourceType'], identifier(_).get('value', None))).filter(lambda _: _[1]).to_list()
    items_without_identifiers = seq(copy2).map(lambda _: (_['resourceType'], identifier(_).get('value', None))).filter(lambda _: not _[1]).to_list()

    return items_with_identifiers, items_without_identifiers


def test_gtex_pipeline(gtex, queries):
    """Query gtex server process using our chain ."""
    items_with_identifiers, items_without_identifiers = retrieve_and_process(gtex, queries)

    assert len(items_with_identifiers) == 1
    assert len(items_without_identifiers) == 1

    assert items_with_identifiers[0][0] == 'ResearchStudy'
    assert items_without_identifiers[0][0] == 'CapabilityStatement'


def test_kf_pipeline(kf, queries):
    """Query KF server,  process using our chain"""
    items_with_identifiers, items_without_identifiers = retrieve_and_process(kf, queries)

    assert len(items_with_identifiers) == 45
    assert len(items_without_identifiers) == 1

    assert items_with_identifiers[0][0] == 'ResearchStudy'
    assert items_without_identifiers[0][0] == 'CapabilityStatement'


def test_dbgap_all_studies(dbgap, queries):
    """Query dbGap server, process using our chain"""
    items_with_identifiers, items_without_identifiers = retrieve_and_process(dbgap, queries)

    assert len(items_with_identifiers) == 2773
    assert len(items_without_identifiers) == 1

    assert items_with_identifiers[0][0] == 'ResearchStudy'
    assert items_without_identifiers[0][0] == 'CapabilityStatement'


def test_dbgap_public_studies(dbgap):
    responses = dbgap.retrieve('ResearchStudy?_security=public,unrestricted&_count=1000')
    counter = Counter([_['resourceType'] for _ in responses])
    print(counter)
    assert counter == {'ResearchStudy': 1893}


def test_dbgap_patient(dbgap):
    # works phs002409
    # does not work phs000635 phs000571 phs001988
    research_study_identifier = 'phs002409'
    # responses = dbgap.retrieve('ResearchSubject?study=' + research_study_identifier)
    responses = dbgap.retrieve(f'Patient?_count=1000&_has:ResearchSubject:individual:study={research_study_identifier}&_revinclude=ResearchSubject:individual')
    counter = Counter([_['resourceType'] for _ in responses])
    print(counter)
    assert counter == {'ResearchSubject': 2, 'Patient': 1}


def test_kf_patient(kf):
    """https://kf-api-fhir-service.kidsfirstdrc.org/Patient?_has:ResearchSubject:individual:study=1883519&_revinclude=Specimen:subject"""
    research_study_identifier = '1883519'
    responses = kf.retrieve(f'Patient?_count=1000&_has:ResearchSubject:individual:study={research_study_identifier}&_revinclude=Specimen:subject&_revinclude=DocumentReference:subject&_revinclude=Condition:subject&_revinclude=MedicationAdministration:subject&_revinclude=Observation:subject&_revinclude=QuestionnaireResponse:subject&_revinclude=MedicationStatement:subject&_revinclude=MedicationDispense:subject')
    counter = Counter([_['resourceType'] for _ in responses])
    print(counter)
    assert counter == {'DocumentReference': 4232, 'Condition': 4109, 'Specimen': 1534, 'Observation': 1199, 'Patient': 753}




from click.testing import CliRunner

from fhir_aggregator.cli import cli


commands = {
    'credentials': ['add', 'delete', 'get', 'list', 'update'],
    'metrics': ['create'],
    'normalize': ['results'],
    'search': ['query'],
    'servers': ['list'],
}


def test_all():
    """Simple test of all commands and subcommands"""
    runner = CliRunner()
    for command_name in commands:
        result = runner.invoke(cli, [command_name, '--help'])
        assert result.exit_code == 0, f'Error: {command_name}'
        for subcommand in commands[command_name]:
            result = runner.invoke(cli, [command_name, subcommand, '--help'])
            assert result.exit_code == 0, f'Error: {command_name} {subcommand}'



import click

from fhir_aggregator.cli import servers_cli, credentials_cli, search_cli, metrics_cli, normalize_cli
from fhir_aggregator.cli.click_util import NaturalOrderGroup


@click.group(cls=NaturalOrderGroup)
def cli():
    """FHIR Aggregator Command Line Interface."""
    pass


# add subcommands to the cli
cli.add_command(servers_cli.cli)
cli.add_command(credentials_cli.cli)
cli.add_command(search_cli.cli)
cli.add_command(metrics_cli.cli)
cli.add_command(normalize_cli.cli)



import itertools
import logging
from abc import abstractmethod
from typing import Callable, Optional, Generator

import inflection
import urllib3.util
from fhirclient import client
from pydantic import BaseModel, model_validator, computed_field
from functional import seq

logger = logging.getLogger(__name__)


LOGGED_ALREADY = set()

def NULL_LAMBDA(*args): None  # noqa


TO_ITERATOR = lambda responses: seq(itertools.chain.from_iterable(responses))  # noqa


class FHIRClient(client.FHIRClient):
    """Instances of this class handle authorizing and talking to FHIR Service.

    Parameters:
        See https://github.com/smart-on-fhir/client-py/blob/master/fhirclient/client.py#L19

    Returns:
        Instance of client, with injected authorization method

    Examples: ::
        from fhir_aggregator.client import FHIRClient
        settings = {
            'app_id': 'my_web_app',
            'api_base': 'https://healthcare.googleapis.com/v1/projects/gcp-testing-308520/locations/us-east4/datasets/testset/fhirStores/fhirstore/fhir'
        }

    """

    def __init__(self, *args, **kwargs):
        """Pass args to super, add authenticator, prepares connection."""
        # grab auth if passed
        auth = None
        if 'auth' in kwargs:
            auth = kwargs['auth']
            del kwargs['auth']

        _settings = kwargs['settings']
        api_base = _settings['api_bases'].pop()
        _settings['api_base'] = api_base

        super(FHIRClient, self).__init__(*args, **kwargs)
        client_major_version = int(client.__version__.split('.')[0])
        assert client_major_version >= 4, f"requires version >= 4.0.0 current version {client.__version__} `pip install -e git+https://github.com/smart-on-fhir/client-py#egg=fhirclient`"
        if auth:
            logger.debug("Setting auth")
            self.server.auth = auth
            self.server.session.hooks['response'].append(self.server.auth.handle_401)
        self.prepare()
        assert self.ready, "server should be ready"

    @staticmethod
    def entries(responses: list, callback: Callable = NULL_LAMBDA) -> Generator[dict, None, None]:
        """Return all resources in a response, by iterating over Bundle.entry."""
        if not isinstance(responses, list):
            responses = [responses]
        for response in responses:
            if response.get('resourceType', '') == 'Bundle' and response.get('total', None) == 0:
                continue
            if 'entry' not in response:
                callback(response)
                yield response
            for _ in response.get('entry', []):
                callback(_['resource'])
                yield _['resource']

    def pages(self, path, callback: Callable = NULL_LAMBDA, fetch_all: bool = True) -> Generator[dict, None, None]:
        """Get all pages of a resource"""
        response = self.server.request_json(path=path)
        yield from self.entries(response)
        while 'link' in response and fetch_all:
            _next = next((lnk for lnk in response['link'] if lnk['relation'] == 'next'), None)
            if not _next:
                break
            next_link = _next['url'].replace(self.server.base_uri, '')
            if 'localhost' in next_link:
                msg = "replacing localhost in next_link"
                if msg not in LOGGED_ALREADY:
                    logger.warning(f"replacing localhost in next_link: {next_link}")
                    LOGGED_ALREADY.add(msg)

                parsed_url = urllib3.util.parse_url(next_link)
                path = parsed_url.path
                query = parsed_url.query
                if query:
                    next_link = f"{path}?{query}"
                else:
                    next_link = path

            # print('NEXT LINK', next_link)
            callback(next_link)
            response = self.server.request_json(path=next_link)
            yield from self.entries(response)

    def retrieve(self, queries, callback: Callable = NULL_LAMBDA, chain: Callable = TO_ITERATOR, fetch_all: bool = True):
        """Query server with queries, internally creates a list of generators, default chain will return an iterator."""
        if not isinstance(queries, list):
            queries = [queries]
        responses = [self.pages(path=_, callback=callback, fetch_all=fetch_all) for _ in queries]
        # return [_ for _ in chain(responses)]
        yield from chain(responses)


class ResourceSummary(BaseModel):
    """All summaries should inherit from this class."""
    warnings: list[str] = []

    @computed_field()
    @property
    def simplified(self) -> dict:
        _ = {'identifier': self.identifier.get('value', None)}
        _.update(self.scalars)
        _.update(self.codings)
        _.update(self.extensions)
        return _

    def simplify_extensions(self, resource: dict = None, _extensions: dict = None) -> dict:
        """Extract extension values, derive key from extension url"""

        def _populate_simplified_extension(_):
            # simple extension
            value_normalized, extension_key = normalize_value(_)
            extension_key = _['url'].split('/')[-1]
            extension_key = inflection.underscore(extension_key).removesuffix(".json").removeprefix("structure_definition_")
            assert value_normalized is not None, f"extension: {extension_key} = {value_normalized} {_}"
            _extensions[extension_key] = value_normalized

        if not _extensions:
            _extensions = {}

        if not resource:
            resource = self.resource_dict

        for _ in resource.get('extension', [resource]):
            if 'extension' not in _.keys():
                if 'resourceType' not in _.keys():
                    _populate_simplified_extension(_)
                continue
            elif set(_.keys()) == set(['url', 'extension']):
                for child_extension in _['extension']:
                    self.simplify_extensions(resource=child_extension, _extensions=_extensions)

        return _extensions

    @property
    @abstractmethod
    def resource_dict(self) -> dict:
        """subclasses should override: Return the resource dictionary."""
        pass

    @computed_field
    @property
    def extensions(self) -> dict:
        return self.simplify_extensions()

    @computed_field
    @property
    def scalars(self) -> dict:
        """Return a dictionary of scalar values."""
        return {k: v for k, v in self.resource_dict.items() if (not isinstance(v, list) and not isinstance(v, dict))}

    @computed_field
    @property
    def codings(self) -> dict:
        """Return a dictionary of scalar values."""
        _codings = {}
        for k, v in self.resource_dict.items():
            if k in ['identifier', 'extension']:
                continue
            if isinstance(v, list):
                for _ in v:
                    if isinstance(_, dict):
                        for value, source in normalize_coding(_):
                            _codings[k] = value
            elif isinstance(v, dict):
                for value, source in normalize_coding(v):
                    _codings[k] = value
        return _codings

    @property
    def identifier(self) -> str:
        """Return the official identifier, or first of a resource."""
        identifiers = self.resource_dict.get('identifier', [])
        official_identifiers = [_ for _ in identifiers if _.get('use', '') == 'official']
        if not official_identifiers and identifiers:
            return identifiers[0]
        elif official_identifiers:
            return official_identifiers[0]
        else:
            return {}


class CapabilityStatementSummary(ResourceSummary):
    """Summarize a capability statement."""
    capability_statement: dict
    _resource_types: list[str] = []

    @property
    def resource_dict(self) -> dict:
        return self.capability_statement

    @model_validator(mode='before')
    def set_summary_fields(cls, values):
        capability_statement = values.get('capability_statement')

        assert capability_statement, f"Expected capability_statement, got {capability_statement}"
        assert isinstance(capability_statement, dict), f"Expected dict, got {type(capability_statement)}"
        assert 'resourceType' in capability_statement, f"Expected resourceType in {capability_statement}"
        assert capability_statement['resourceType'] == 'CapabilityStatement', f"Expected CapabilityStatement, got {capability_statement['resourceType']}"

        if 'rest' not in capability_statement:
            values['warnings'] = []
            values['warnings'].append(f"CapabilityStatement has no rest")
        return values

    @property
    def resource_types(self) -> list[str]:
        resource_types = []
        capability_statement = self.capability_statement
        if not self._resource_types:
            if 'rest' not in capability_statement:
                logger.warning(f"CapabilityStatement has no rest: {capability_statement}")
            else:
                for rest in capability_statement['rest']:
                    if 'resource' not in rest:
                        logger.warning(f"Rest has no resource: {rest}")
                    else:
                        for resource in rest['resource']:
                            resource_types.append(resource['type'])
            self._resource_types = sorted(resource_types)
        return self._resource_types

    @property
    def everything(self) -> list[str]:
        assert self.capability_statement, f"Expected capability_statement, got {self.capability_statement}"
        assert 'rest' in self.capability_statement, f"Expected rest in {self.capability_statement}"
        for rest in self.capability_statement['rest']:
            assert 'operation' in rest, f"Expected operation in {rest}"
            for operation in rest['operation']:
                _name = operation['name']


class ResearchStudySummary(ResourceSummary):
    """Summarize a research study."""

    research_study: dict

    @computed_field
    @property
    def resource_dict(self) -> dict:
        return self.research_study


class OperationDefinitionSummary(ResourceSummary):
    """Summarize an operation definition."""

    operation_definition: list[dict]

    @property
    def resource_dict(self) -> dict:
        return self.operation_definition[0]

    @property
    def everything(self) -> list[str]:
        """Resources that have an $everything operation."""
        resources = []
        for _ in self.operation_definition:
            if _['code'] != 'everything':
                continue
            resources.extend(_['resource'])
        return sorted(resources)


def normalize_value(resource_dict: dict) -> tuple[Optional[str], Optional[str]]:
    """return a tuple containing the normalized value and the name of the field it was derived from"""
    value_normalized = None
    value_source = None

    if 'valueQuantity' in resource_dict:
        value = resource_dict['valueQuantity']
        value_normalized = f"{value['value']} {value.get('unit', '')}"
        value_source = 'valueQuantity'
    elif 'valueCodeableConcept' in resource_dict:
        value = resource_dict['valueCodeableConcept']
        value_normalized = ' '.join([coding.get('display', coding.get('code', '')) for coding in value.get('coding', [])])
        value_source = 'valueCodeableConcept'
    elif 'valueCoding' in resource_dict:
        value = resource_dict['valueCoding']
        value_normalized = value['display']
        value_source = 'valueCoding'
    elif 'valueString' in resource_dict:
        value_normalized = resource_dict['valueString']
        value_source = 'valueString'
    elif 'valueCode' in resource_dict:
        value_normalized = resource_dict['valueCode']
        value_source = 'valueCode'
    elif 'valueBoolean' in resource_dict:
        value_normalized = str(resource_dict['valueBoolean'])
        value_source = 'valueBoolean'
    elif 'valueInteger' in resource_dict:
        value_normalized = str(resource_dict['valueInteger'])
        value_source = 'valueInteger'
    elif 'valueRange' in resource_dict:
        value = resource_dict['valueRange']
        low = value['low']
        high = value['high']
        value_normalized = f"{low['value']} - {high['value']} {low.get('unit', '')}"
        value_source = 'valueRange'
    elif 'valueRatio' in resource_dict:
        value = resource_dict['valueRatio']
        numerator = value['numerator']
        denominator = value['denominator']
        value_normalized = f"{numerator['value']} {numerator.get('unit', '')}/{denominator['value']} {denominator.get('unit', '')}"
        value_source = 'valueRatio'
    elif 'valueSampledData' in resource_dict:
        value = resource_dict['valueSampledData']
        value_normalized = value['data']
        value_source = 'valueSampledData'
    elif 'valueTime' in resource_dict:
        value_normalized = resource_dict['valueTime']
        value_source = 'valueTime'
    elif 'valueDateTime' in resource_dict:
        value_normalized = resource_dict['valueDateTime']
        value_source = 'valueDateTime'
    elif 'valuePeriod' in resource_dict:
        value = resource_dict['valuePeriod']
        value_normalized = f"{value['start']} to {value['end']}"
        value_source = 'valuePeriod'
    elif 'valueUrl' in resource_dict:
        value_normalized = resource_dict['valueUrl']
        value_source = 'valueUrl'
    elif 'valueDate' in resource_dict:
        value_normalized = resource_dict['valueDate']
        value_source = 'valueDate'
    elif 'valueCount' in resource_dict:
        value_normalized = resource_dict['valueCount']['value']
        value_source = 'valueCount'
    # for debugging...
    else:
        raise ValueError(f"value[x] not found in {resource_dict}")

    return value_normalized, value_source


def normalize_coding(resource_dict: dict) -> list[tuple[str, str]]:
    def extract_coding(coding_list):
        # return a concatenated string
        # return ','.join([coding.get('display', '') for coding in coding_list if 'display' in coding])
        # or alternatively return an array
        return [coding.get('display', coding.get('code', '')) for coding in coding_list]

    def find_codings_in_dict(d: dict, parent_key: str = '') -> list[tuple[str, str]]:
        codings = []
        for key, value in d.items():
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        # Check if the dict contains a 'coding' list
                        if 'coding' in item and isinstance(item['coding'], list):
                            coding_string = extract_coding(item['coding'])
                            codings.append((coding_string, key))
                        if 'code' in item:
                            # print(f">>> DBG value: {item} {key} {parent_key}")
                            coding_string = item['code']
                            codings.append((coding_string, key))

                        # Recursively search in the dict
                        codings.extend(find_codings_in_dict(item, key))
            elif isinstance(value, dict):
                # Check if the dict contains a 'coding' list
                if 'coding' in value and isinstance(value['coding'], list):
                    coding_string = extract_coding(value['coding'])
                    codings.append((coding_string, key))

                # Recursively search in the dict
                codings.extend(find_codings_in_dict(value, key))
        return codings

    return find_codings_in_dict(resource_dict)


def is_number(s):
    """ Returns True if string is a number. """
    try:
        complex(s)
        return True
    except ValueError:
        return False





import itertools
from collections import Counter
from unittest.mock import patch

import pytest
import requests
from fhirclient.server import FHIRUnauthorizedException
from functional import seq

from fhir_aggregator.client import ResearchStudySummary
from pprint import pprint


@pytest.fixture()
def unauthorized_request(mocker):
    mocker.status_code = 401
    mocker.json = lambda: {'error': 'Unauthorized'}
    return mocker


def test_unauthorized_request(dbgap, unauthorized_request):
    with pytest.raises(FHIRUnauthorizedException):
        with patch.object(requests.Session, 'get', return_value=unauthorized_request):
            responses = dbgap.retrieve(['metadata'])
            assert seq(responses).first()


def test_dbgap(dbgap):
    """Query dbGap server."""
    responses = dbgap.retrieve('ResearchStudy/phs002409')
    research_study = seq(responses).first()
    summary = ResearchStudySummary(research_study=research_study)

    print(summary.identifier)
    assert summary.identifier['value'] == 'phs002409.v1.p1'

    print(summary.extensions)
    assert summary.extensions == {
        'research_study_study_overview_url': 'https://www.ncbi.nlm.nih.gov/projects/gap/cgi-bin/study.cgi?study_id=phs002409.v1.p1',
        'research_study_release_date': '2021-09-09', 'research_study_study_consents_study_consent': 'GRU',
        'research_study_content_num_phenotype_datasets': 2, 'research_study_content_num_molecular_datasets': 0,
        'research_study_content_num_variables': 58, 'research_study_content_num_documents': 0,
        'research_study_content_num_analyses': 0, 'research_study_content_num_subjects': 813,
        'research_study_content_num_samples': 0, 'research_study_content_num_sub_studies': 0}

    pprint(summary.research_study['extension'])

    responses = dbgap.retrieve('Patient?_count=1000&_has:ResearchSubject:individual:study=' + summary.research_study['id'])
    copy1, copy2 = itertools.tee(responses, 2)
    counter = Counter([_['resourceType'] for _ in copy1])
    print(counter)
    assert counter == {'Patient': 813}

    responses = dbgap.retrieve(f'Patient/{seq(copy2).first()["id"]}')
    patient = seq(responses).first()
    assert patient
    print(patient)
    assert patient['resourceType'] == 'Patient'



from collections import OrderedDict

import click


# https://github.com/pallets/click/issues/513#issuecomment-301046782
class NaturalOrderGroup(click.Group):
    """Command group trying to list subcommands in the order they were added.
    Example use::

    @click.group(cls=NaturalOrderGroup)

    If passing dict of commands from other sources, ensure they are of type
    OrderedDict and properly ordered, otherwise order of them will be random
    and newly added will come to the end.
    """

    def __init__(self, name=None, commands=None, **attrs):
        if commands is None:
            commands = OrderedDict()
        elif not isinstance(commands, OrderedDict):
            commands = OrderedDict(commands)
        click.Group.__init__(self, name=name,
                             commands=commands,
                             **attrs)

    def list_commands(self, ctx):
        """List command names as they are in commands dict.

        If the dict is OrderedDict, it will preserve the order commands
        were added.
        """
        return self.commands.keys()



import pytest

from fhir_aggregator import Credential


def test_credential_no_endpoint():
    """Ensure endpoint checked."""
    with pytest.raises(ValueError):
        Credential(endpoint=None, token='token', username='username', password='password')
    Credential(endpoint="endpoint", token='token', username='username', password='password')


def test_credential_no_creds():
    """Ensure creds checked."""
    with pytest.raises(ValueError):
        Credential(endpoint="endpoint",)
    Credential(endpoint="endpoint", token="token")
    Credential(endpoint="endpoint",  username='username', password='password')


import json

import pytest
from pydantic import ValidationError

from fhir_aggregator.client import CapabilityStatementSummary


@pytest.fixture()
def dbgap_capability_statement(mocker):
    with open('tests/fixtures/dbgap_capability_statement.json') as fp:
        _ = json.load(fp)
    mocker.retrieve = lambda queries: _
    return mocker


@pytest.fixture()
def kf_capability_statement(mocker):
    with open('tests/fixtures/kf_capability_statement.json') as fp:
        _ = json.load(fp)
    mocker.retrieve = lambda queries: _
    return mocker


@pytest.fixture()
def gtex_capability_statement(mocker):
    with open('tests/fixtures/gtex_capability_statement.json') as fp:
        _ = json.load(fp)
    mocker.retrieve = lambda queries: _
    return mocker


def test_capability_statement_summary():
    """Test the capability statement summary, basic validation."""
    with pytest.raises(ValidationError):
        CapabilityStatementSummary(capability_statement=None)
    with pytest.raises(ValidationError):
        CapabilityStatementSummary(capability_statement={})
    with pytest.raises(ValidationError):
        CapabilityStatementSummary(capability_statement={'resourceType': 'Foo'})

    capability_statement_summary = CapabilityStatementSummary(capability_statement={'resourceType': 'CapabilityStatement'})
    assert len(capability_statement_summary.warnings) > 0


def test_dbgap_capability_statement(dbgap_capability_statement):
    """Query dbGap server."""
    capability_statement = dbgap_capability_statement.retrieve(['metadata'])
    assert capability_statement
    assert isinstance(capability_statement, dict)
    dbgap_capability_statement_summary = CapabilityStatementSummary(capability_statement=capability_statement)
    assert dbgap_capability_statement_summary.resource_types == ['Observation', 'OperationDefinition', 'Patient', 'ResearchStudy']
    assert len(dbgap_capability_statement_summary.extensions) == 0


def test_kf_capability_statement(kf_capability_statement):
    """Query kf server."""
    capability_statement = kf_capability_statement.retrieve(['metadata'])
    assert capability_statement
    assert isinstance(capability_statement, dict)
    kf_capability_statement_summary = CapabilityStatementSummary(capability_statement=capability_statement)
    print(kf_capability_statement_summary.resource_types)
    assert kf_capability_statement_summary.resource_types == ['Account', 'ActivityDefinition', 'AdverseEvent',
                                                              'AllergyIntolerance', 'Appointment',
                                                              'AppointmentResponse', 'AuditEvent', 'Basic', 'Binary',
                                                              'BiologicallyDerivedProduct', 'BodyStructure', 'Bundle',
                                                              'CapabilityStatement', 'CarePlan', 'CareTeam',
                                                              'CatalogEntry', 'ChargeItem', 'ChargeItemDefinition',
                                                              'Claim', 'ClaimResponse', 'ClinicalImpression',
                                                              'CodeSystem', 'Communication', 'CommunicationRequest',
                                                              'CompartmentDefinition', 'Composition', 'ConceptMap',
                                                              'Condition', 'Consent', 'Contract', 'Coverage',
                                                              'CoverageEligibilityRequest',
                                                              'CoverageEligibilityResponse', 'DetectedIssue', 'Device',
                                                              'DeviceDefinition', 'DeviceMetric', 'DeviceRequest',
                                                              'DeviceUseStatement', 'DiagnosticReport',
                                                              'DocumentManifest', 'DocumentReference',
                                                              'EffectEvidenceSynthesis', 'Encounter', 'Endpoint',
                                                              'EnrollmentRequest', 'EnrollmentResponse',
                                                              'EpisodeOfCare', 'EventDefinition', 'Evidence',
                                                              'EvidenceVariable', 'ExampleScenario',
                                                              'ExplanationOfBenefit', 'FamilyMemberHistory', 'Flag',
                                                              'Goal', 'GraphDefinition', 'Group', 'GuidanceResponse',
                                                              'HealthcareService', 'ImagingStudy', 'Immunization',
                                                              'ImmunizationEvaluation', 'ImmunizationRecommendation',
                                                              'ImplementationGuide', 'InsurancePlan', 'Invoice',
                                                              'Library', 'Linkage', 'List', 'Location', 'Measure',
                                                              'MeasureReport', 'Media', 'Medication',
                                                              'MedicationAdministration', 'MedicationDispense',
                                                              'MedicationKnowledge', 'MedicationRequest',
                                                              'MedicationStatement', 'MedicinalProduct',
                                                              'MedicinalProductAuthorization',
                                                              'MedicinalProductContraindication',
                                                              'MedicinalProductIndication',
                                                              'MedicinalProductIngredient',
                                                              'MedicinalProductInteraction',
                                                              'MedicinalProductManufactured',
                                                              'MedicinalProductPackaged',
                                                              'MedicinalProductPharmaceutical',
                                                              'MedicinalProductUndesirableEffect', 'MessageDefinition',
                                                              'MessageHeader', 'MolecularSequence', 'NamingSystem',
                                                              'NutritionOrder', 'Observation', 'ObservationDefinition',
                                                              'OperationDefinition', 'OperationOutcome', 'Organization',
                                                              'OrganizationAffiliation', 'Parameters', 'Patient',
                                                              'PaymentNotice', 'PaymentReconciliation', 'Person',
                                                              'PlanDefinition', 'Practitioner', 'PractitionerRole',
                                                              'Procedure', 'Provenance', 'Questionnaire',
                                                              'QuestionnaireResponse', 'RelatedPerson', 'RequestGroup',
                                                              'ResearchDefinition', 'ResearchElementDefinition',
                                                              'ResearchStudy', 'ResearchSubject', 'RiskAssessment',
                                                              'RiskEvidenceSynthesis', 'Schedule', 'SearchParameter',
                                                              'ServiceRequest', 'Slot', 'Specimen',
                                                              'SpecimenDefinition', 'StructureDefinition',
                                                              'StructureMap', 'Subscription', 'Substance',
                                                              'SubstanceNucleicAcid', 'SubstancePolymer',
                                                              'SubstanceProtein', 'SubstanceReferenceInformation',
                                                              'SubstanceSourceMaterial', 'SubstanceSpecification',
                                                              'SupplyDelivery', 'SupplyRequest', 'Task',
                                                              'TerminologyCapabilities', 'TestReport', 'TestScript',
                                                              'ValueSet', 'VerificationResult', 'VisionPrescription']
    assert len(kf_capability_statement_summary.extensions) == 0


def test_gtex_capability_statement(gtex_capability_statement):
    """Query kf server."""
    capability_statement = gtex_capability_statement.retrieve(['metadata'])
    assert capability_statement
    assert isinstance(capability_statement, dict)
    gtex_capability_statement_summary = CapabilityStatementSummary(capability_statement=capability_statement)
    print(gtex_capability_statement_summary.resource_types)
    assert gtex_capability_statement_summary.resource_types == ['Account', 'ActivityDefinition', 'AdverseEvent',
                                                                'AllergyIntolerance', 'Appointment',
                                                                'AppointmentResponse', 'AuditEvent', 'Basic', 'Binary',
                                                                'BiologicallyDerivedProduct', 'BodyStructure', 'Bundle',
                                                                'CapabilityStatement', 'CarePlan', 'CareTeam',
                                                                'CatalogEntry', 'ChargeItem', 'ChargeItemDefinition',
                                                                'Claim', 'ClaimResponse', 'ClinicalImpression',
                                                                'CodeSystem', 'Communication', 'CommunicationRequest',
                                                                'CompartmentDefinition', 'Composition', 'ConceptMap',
                                                                'Condition', 'Consent', 'Contract', 'Coverage',
                                                                'CoverageEligibilityRequest',
                                                                'CoverageEligibilityResponse', 'DetectedIssue',
                                                                'Device', 'DeviceDefinition', 'DeviceMetric',
                                                                'DeviceRequest', 'DeviceUseStatement',
                                                                'DiagnosticReport', 'DocumentManifest',
                                                                'DocumentReference', 'DomainResource',
                                                                'EffectEvidenceSynthesis', 'Encounter', 'Endpoint',
                                                                'EnrollmentRequest', 'EnrollmentResponse',
                                                                'EpisodeOfCare', 'EventDefinition', 'Evidence',
                                                                'EvidenceVariable', 'ExampleScenario',
                                                                'ExplanationOfBenefit', 'FamilyMemberHistory', 'Flag',
                                                                'Goal', 'GraphDefinition', 'Group', 'GuidanceResponse',
                                                                'HealthcareService', 'ImagingStudy', 'Immunization',
                                                                'ImmunizationEvaluation', 'ImmunizationRecommendation',
                                                                'ImplementationGuide', 'InsurancePlan', 'Invoice',
                                                                'Library', 'Linkage', 'List', 'Location', 'Measure',
                                                                'MeasureReport', 'Media', 'Medication',
                                                                'MedicationAdministration', 'MedicationDispense',
                                                                'MedicationKnowledge', 'MedicationRequest',
                                                                'MedicationStatement', 'MedicinalProduct',
                                                                'MedicinalProductAuthorization',
                                                                'MedicinalProductContraindication',
                                                                'MedicinalProductIndication',
                                                                'MedicinalProductIngredient',
                                                                'MedicinalProductInteraction',
                                                                'MedicinalProductManufactured',
                                                                'MedicinalProductPackaged',
                                                                'MedicinalProductPharmaceutical',
                                                                'MedicinalProductUndesirableEffect',
                                                                'MessageDefinition', 'MessageHeader',
                                                                'MolecularSequence', 'NamingSystem', 'NutritionOrder',
                                                                'Observation', 'ObservationDefinition',
                                                                'OperationDefinition', 'OperationOutcome',
                                                                'Organization', 'OrganizationAffiliation', 'Parameters',
                                                                'Patient', 'PaymentNotice', 'PaymentReconciliation',
                                                                'Person', 'PlanDefinition', 'Practitioner',
                                                                'PractitionerRole', 'Procedure', 'Provenance',
                                                                'Questionnaire', 'QuestionnaireResponse',
                                                                'RelatedPerson', 'RequestGroup', 'ResearchDefinition',
                                                                'ResearchElementDefinition', 'ResearchStudy',
                                                                'ResearchSubject', 'Resource', 'RiskAssessment',
                                                                'RiskEvidenceSynthesis', 'Schedule', 'SearchParameter',
                                                                'ServiceRequest', 'Slot', 'Specimen',
                                                                'SpecimenDefinition', 'StructureDefinition',
                                                                'StructureMap', 'Subscription', 'Substance',
                                                                'SubstanceNucleicAcid', 'SubstancePolymer',
                                                                'SubstanceProtein', 'SubstanceReferenceInformation',
                                                                'SubstanceSourceMaterial', 'SubstanceSpecification',
                                                                'SupplyDelivery', 'SupplyRequest', 'Task',
                                                                'TerminologyCapabilities', 'TestReport', 'TestScript',
                                                                'ValueSet', 'VerificationResult', 'VisionPrescription']
    assert len(gtex_capability_statement_summary.extensions) == 0



import json

import pytest
from pydantic import ValidationError

from fhir_aggregator.client import CapabilityStatementSummary, OperationDefinitionSummary


@pytest.fixture()
def dbgap_operation_definition(mocker):
    with open('tests/fixtures/dbgap_operation_definition.json') as fp:
        _ = json.load(fp)
    mocker.retrieve = lambda queries: _
    return mocker


@pytest.fixture()
def kf_operation_definition(mocker):
    with open('tests/fixtures/kf_operation_definition.json') as fp:
        _ = json.load(fp)
    mocker.retrieve = lambda queries: _
    return mocker


@pytest.fixture()
def gtex_operation_definition(mocker):
    with open('tests/fixtures/gtex_operation_definition.json') as fp:
        _ = json.load(fp)
    mocker.retrieve = lambda queries: _
    return mocker


def test_dbgap_operation_definition(dbgap_operation_definition):
    """Query dbGap server."""
    #  Note dbgap does not return OperationDefinition
    operation_definition = dbgap_operation_definition.retrieve(['OperationDefinition?_count=1000'])
    assert isinstance(operation_definition, list)
    operation_definition_summary = OperationDefinitionSummary(operation_definition=operation_definition)
    assert operation_definition_summary.everything == []


def test_kf_operation_definition(kf_operation_definition):
    """Query kf server."""
    operation_definition = kf_operation_definition.retrieve(['OperationDefinition?_count=1000'])
    assert operation_definition
    assert isinstance(operation_definition, list)
    operation_definition_summary = OperationDefinitionSummary(operation_definition=operation_definition)
    assert operation_definition_summary
    print(operation_definition_summary.everything)
    assert operation_definition_summary.everything == ['Encounter', 'Group', 'MedicinalProduct', 'Patient']


def test_gtex_operation_definition(gtex_operation_definition):
    """Query gtex server."""
    #  Note gtex does not return OperationDefinition
    operation_definition = gtex_operation_definition.retrieve(['OperationDefinition?_count=1000'])
    assert isinstance(operation_definition, list)
    operation_definition_summary = OperationDefinitionSummary(operation_definition=operation_definition)
    assert operation_definition_summary.everything == []



from setuptools import setup, find_packages

# Step 1: Read requirements.txt for dependencies
with open('requirements.txt') as f:
    required = f.read().splitlines()

# Step 2: Read README.md for the long description
with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

# Step 3: Define the setup
setup(
    name='fhir-aggregator',
    version='0.2rc2',
    packages=find_packages(),
    install_requires=required,
    long_description=long_description,
    long_description_content_type='text/markdown',
    entry_points={
        'console_scripts': [
            'fhir-aggregator=fhir_aggregator.cli:cli',
        ],
    },
    # Add more package metadata as needed
)



import pytest

from tests.integration import SERVERS


@pytest.fixture()
def gtex():
    return SERVERS['gtex']


@pytest.fixture()
def dbgap():
    return SERVERS['dbgap']


@pytest.fixture()
def kf():
    return SERVERS['kf']


@pytest.fixture()
def queries():
    """Queries to run on FHIR servers."""
    return [
        'metadata',
        'ResearchStudy?_count=1000',
    ]



import click
from fhir_aggregator import CredentialManager, Credential
from fhir_aggregator.cli.click_util import NaturalOrderGroup


@click.group(name='credentials', cls=NaturalOrderGroup)
def cli():
    """Credential management commands."""
    pass


@cli.command()
@click.argument('endpoint')
@click.option('--token', default=None, help='Authentication token')
@click.option('--username', default=None, help='Username for authentication')
@click.option('--password', default=None, help='Password for authentication')
def add(endpoint, token, username, password):
    """Add a new credential."""
    manager = CredentialManager()
    credential = Credential(endpoint=endpoint, token=token, username=username, password=password)
    manager.add_or_update_credential(credential)
    click.echo(f'Credential for {endpoint} added.')


@cli.command()
@click.argument('endpoint')
def delete(endpoint):
    """Delete a credential."""
    manager = CredentialManager()
    manager.delete_credential(endpoint)
    click.echo(f'Credential for {endpoint} deleted.')


@cli.command()
@click.argument('endpoint')
def get(endpoint):
    """Get a credential."""
    manager = CredentialManager()
    credential = manager.get_credential(endpoint)
    if credential:
        click.echo(f'Credential: {credential}')
    else:
        click.echo('Credential not found.')


@cli.command(name='list')
def _list():
    """List all credentials."""
    manager = CredentialManager()
    credentials = manager.list_credentials()  # Assuming this method exists
    for endpoint, credential in credentials.items():
        click.echo(f'{endpoint}: {credential}')


@cli.command()
@click.argument('endpoint')
@click.option('--token', default=None, help='Authentication token')
@click.option('--username', default=None, help='Username for authentication')
@click.option('--password', default=None, help='Password for authentication')
def update(endpoint, token, username, password):
    """Update an existing credential."""
    manager = CredentialManager()
    credential = Credential(endpoint=endpoint, token=token, username=username, password=password)
    manager.add_or_update_credential(credential)
    click.echo(f'Credential for {endpoint} updated.')



import click
import yaml

from fhir_aggregator import search_all_servers
from fhir_aggregator.cli.click_util import NaturalOrderGroup


@click.group(name='search', cls=NaturalOrderGroup)
def cli():
    """Search commands."""
    pass


@cli.command()
@click.argument('yaml_file', type=click.Path(exists=True))
def query(yaml_file):
    """
    Reads a YAML file of queries and searches all servers.

    :param yaml_file: Path to the YAML file containing the queries.
    """
    with open(yaml_file, 'r') as file:
        queries = yaml.safe_load(file)

    # Assuming queries is a list of query strings
    for _ in queries:
        results = search_all_servers(_)
        # Process or display the results
        click.echo(f'Results for query "{_}": {results}')





from fhir_aggregator.client import FHIRClient
from fhir_aggregator.client.smart_auth import GoogleFHIRAuth, DBGapFHIRAuth

# set up servers for testing
SERVERS = {}

project = 'ncpi-fhir-cat-2022/locations/us-central1/datasets/GTEx_Open_Access/fhirStores/gtex_v8.2/fhir/'
settings = {
    'app_id': 'gtex',
    'api_bases': [
        'https://healthcare.googleapis.com/v1/projects/' + project,
    ]
}

SERVERS['gtex'] = FHIRClient(settings=settings, auth=GoogleFHIRAuth())

"""Connect to open Kids First fhir server, with several datasets"""
settings = {
    'app_id': 'kf',
    'api_bases': [
        'https://kf-api-fhir-service.kidsfirstdrc.org/',
    ]
}

SERVERS['kf'] = FHIRClient(settings=settings)

settings = {
    'app_id': 'dbgap',
    'api_bases': [
        'https://dbgap-api.ncbi.nlm.nih.gov/fhir/x1',
    ]
}

SERVERS['dbgap'] = FHIRClient(settings=settings, auth=DBGapFHIRAuth(path='tests/fixtures/dbgap-task-specific-token.txt'))
